"""Copyright 2021 Google LLC.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.

You may obtain a copy of the License at
                https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

---

Code taken and adapted from

  https://github.com/google/gnxi/tree/master/gnmi_cli_py, under
  Apache License, Version 2.0 license
"""


import json
import logging
import ssl
import time
from typing import Any, List, Optional, Tuple

import grpc
from grpc._channel import _InactiveRpcError, _MultiThreadedRendezvous

from oc_config_validate import context, schema
from oc_config_validate.gnmi import gnmi_pb2, gnmi_pb2_grpc  # type: ignore


class BaseError(Exception):
    """Base Error for the target class"""


class TargetError(BaseError):
    """Error connecting to Target."""


class RpcError(BaseError):
    """Error on the RPC."""


class TestTarget():
    """gNMI Target to run tests against.

    Example:
        with target.TestTarget(ctx.tgt) as tgt:
            tgt.gNMISet()
            tgt.gNMIGet()
    """

    def __init__(self, tgt: context.Target):
        self.address = tgt.target
        self.username = tgt.username
        self.password = tgt.password
        self.notls = tgt.no_tls
        self.host_tls_override = tgt.tls_host_override
        self.target_cert_as_root_ca = tgt.target_cert_as_root_ca
        self.gnmi_set_cooldown_secs = tgt.gnmi_set_cooldown_secs
        self.key = None
        self.cert = None
        self.root_ca = None
        self.channel = None
        self.stub = None

        if not self.notls:
            self.setCertificates(key=tgt.private_key, cert=tgt.cert_chain,
                                 root_ca=tgt.root_ca_cert)

    def __repr__(self):
        return ('TestTarget(address=%r, username=%r, notls=%r)' %
                (self.address, self.username, self.notls))

    def __str__(self):
        return self.address

    def setCertificates(self, key: Optional[str] = None,
                        cert: Optional[str] = None,
                        root_ca: Optional[str] = None,):
        """Read the TLS certificates to use on connections.

        Args:
            key: Path to the private key file.
                 If None, no key will be used.
            cert: Path to the certificate to present to the target.
                  If None, no certificate will be used.
            root_ca: Path to the trusted CA certificate. If None, it will
                     be requested from the Target.
        """
        def _fileReadOrNone(filename: str) -> Optional[bytes]:
            try:
                with open(filename, 'rb') as file:
                    return file.read()
            except IOError as io_error:
                logging.error("Unable to read %s: %s", filename, io_error)
                return None
        if key:
            self.key = _fileReadOrNone(key)
        if cert:
            self.cert = _fileReadOrNone(cert)
        if root_ca:
            self.root_ca = _fileReadOrNone(root_ca)

    def gNMIConnect(self):
        """Create a gNMI connection to the target."""
        if self.notls:
            self.channel = grpc.insecure_channel(self.address)
        else:
            creds = self._buildCredentials()
            logging.debug("creds: %s", creds)
            if self.host_tls_override:
                self.channel = grpc.secure_channel(
                    self.address, creds,
                    (('grpc.ssl_target_name_override',
                      self.host_tls_override,),))
            else:
                self.channel = grpc.secure_channel(
                    self.address, creds)
        self.stub = gnmi_pb2_grpc.gNMIStub(self.channel)

    def gNMIClose(self):
        """Closes the gNMI connection to the target."""
        self.channel.close()
        self.stub = None

    def __enter__(self):
        self.gNMIConnect()
        return self

    def __exit__(self, type, value, traceback):
        self.gNMIClose()

    def _buildCredentials(self) -> grpc.ssl_channel_credentials:
        """Build credentials used in gNMI Requests.

        Returns:
            A gRPC credentials object.

        Raises:
            TargetError if unable to build the credentials.
        """
        if self.target_cert_as_root_ca:
            logging.info('Obtaining Root CA certificate from Target')
            address_elems = self.address.split(':')
            try:
                self.root_ca = ssl.get_server_certificate(
                    (address_elems[0], int(address_elems[1]))).encode('utf-8')
            except ConnectionRefusedError as err:
                raise TargetError(
                    "Unable to get Root CA from Target") from err

        return grpc.ssl_channel_credentials(
            root_certificates=self.root_ca,
            private_key=self.key,
            certificate_chain=self.cert)

    def _GnmiAuthMetadata(self) -> List[Tuple[str, str]]:
        """Returns the gNMI metadata used for authentication."""
        if self.username:  # User/pass supplied for Authentication.
            return [('username', self.username), ('password', self.password)]
        return []

    def gNMIGet(self, xpath: str) -> gnmi_pb2.GetResponse:
        """Create a gNMI GetRequest.

        Args:
            xpath: gNMI Path

        Returns:
            A gnmi_pb2.GetResponse object.

        Raises:
            RpcError if unable to connect to Target.
        """
        path = schema.parsePath(xpath)
        try:
            return self.stub.Get(
                gnmi_pb2.GetRequest(path=[path], encoding='JSON_IETF'),
                metadata=self._GnmiAuthMetadata())
        except _InactiveRpcError as err:
            raise RpcError(err) from err

    def _gNMISet(self, xpath: str, set_type: str,
                 value: Optional[Any] = None) -> gnmi_pb2.SetResponse:
        """Create a gNMI SetRequest.

        Args:
            xpath: The gNMI path to set.
            set_type: UPDATE | DELETE | REPLACE
            value: Value to set; can be numeric, string or JSON-IETF.

        Returns:
            A gnmi_pb2.SetResponse object.

        Raises:
            RpcError if unable to connect to Target.
        """
        if set_type.lower() not in ['delete', 'update', 'replace']:
            raise RpcError('%s is not a valid gNMI Set type' % set_type)

        path = schema.parsePath(xpath)
        path_val = None
        if value:
            val = schema.pythonToTypedValue(value)
            path_val = gnmi_pb2.Update(path=path, val=val)

        auth = self._GnmiAuthMetadata()
        response = None
        try:
            if set_type.lower() == 'delete':
                response = self.stub.Set(gnmi_pb2.SetRequest(delete=[path]),
                                         metadata=auth)
            if set_type.lower() == 'update':
                response = self.stub.Set(gnmi_pb2.SetRequest(
                    update=[path_val]), metadata=auth)
            if set_type.lower() == 'replace':
                response = self.stub.Set(gnmi_pb2.SetRequest(
                    replace=[path_val]), metadata=auth)
        except _InactiveRpcError as err:
            raise RpcError(err) from err

        if self.gnmi_set_cooldown_secs:
            time.sleep(self.gnmi_set_cooldown_secs)

        return response

    def gNMISetUpdate(self, xpath: str,
                      value: Any) -> gnmi_pb2.SetResponse:
        """Create a gNMI SetRequest Update.

        Args:
            xpath: The gNMI path to replace.
            value: Value to set; can be numeric, string or JSON-IETF.

        Returns:
            A gnmi_pb2.SetResponse object.
        """
        return self._gNMISet(xpath, 'update', value)

    def gNMISetReplace(self, xpath: str,
                       value: Any) -> gnmi_pb2.SetResponse:
        """Create a gNMI SetRequest Replace.

        Args:
            xpath: The gNMI path to replace.
            value: Value to set; can be numeric, string or JSON-IETF.

        Returns:
            A gnmi_pb2.SetResponse object.
        """
        return self._gNMISet(xpath, 'replace', value)

    def gNMISetDelete(self, xpath: str) -> gnmi_pb2.SetResponse:
        """Create a gNMI SetRequest Delete.

        Args:
            xpath: The gNMI path to delete.

        Returns:
            A gnmi_pb2.SetResponse object.
        """
        return self._gNMISet(xpath, 'delete')

    def gNMISetConfigFile(self, file_path: str, xpath: str):
        """Apply the JSON-formated configuration to the target.

        Args:
            file_path: Path to a JSON file with the configuration to apply.
            xpath: gNMI xpath where to apply the configuration.
        """
        with open(file_path, encoding="utf8") as file:
            json_data = json.load(file)
        self.gNMISetUpdate(xpath, json_data)

    def _gNMISubscribe(self,
                       request: gnmi_pb2.SubscribeRequest,
                       timeout: int = 30
                       ) -> List[gnmi_pb2.Notification]:
        """Subscribes up to a timeout, returns the Notifications received.

        All Updates received are accumulated and returned when the channel is
         closed (by server, by timeout or error).

        Args:
            request: gNMI Subscription request to set.
            timeout: Seconds to keep the gRPC channel open.

        Returns:
            A list of gnmi_pb2.Notification objects received.

        Raises:
            RpcError if unable to connect to Target.
            ValueError if the SubscribeResponse is invalid.
            GnmiError if no sync_response was received.
        """

        auth = self._GnmiAuthMetadata()
        notifications = []
        got_sync_response = False
        try:
            for resp in self.stub.Subscribe(
                    iter([request]), timeout=timeout, metadata=auth):
                if resp.sync_response:
                    got_sync_response = True
                elif resp.update:
                    notifications.append(resp.update)
                else:
                    raise ValueError("Invalid SubscribeResponse %s" % resp)
        except _MultiThreadedRendezvous as err:
            if err.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
                return notifications
            else:
                raise RpcError(err) from err
        except _InactiveRpcError as err:
            raise RpcError(err) from err
        if not got_sync_response:
            raise schema.GnmiError(
                "No Response with sync_response was received")
        return notifications

    def gNMISubsOnce(self, xpaths: List[str]) -> List[gnmi_pb2.Notification]:
        """Subscribes using ONCE mode, returns the Notifications received.

        Args:
            xpaths: List of gNMI paths to subscribe to.

        Returns:
            A list of gnmi_pb2.Notification objects received.
        """
        request = schema.gNMISubscriptionOnceRequest(xpaths)
        return self._gNMISubscribe(request)

    def gNMISubsStreamSample(self, xpath: str, sample_interval: int,
                             timeout: int) -> List[gnmi_pb2.Notification]:
        """Subscribes using STREAM mode, returns the Notifications received.

        Args:
            xpath: gNMI path to subscribe to.
            sample_interval: Nanoseconds between updates.
            timeout: Seconds to keep the gRPC channel open and receive
                updates.

        Returns:
            A list of gnmi_pb2.Notification objects received.
        """
        request = schema.gNMISubscriptionStreamSampleRequest(
            [xpath], sample_interval)
        return self._gNMISubscribe(request, timeout=timeout)

    def gNMISubsStreamOnChange(
            self,
            xpath: str,
            timeout: int = 30,
            sync_response_callback=None) -> Tuple[
                List[gnmi_pb2.Notification],
                List[gnmi_pb2.Notification]]:
        """Subscribes on-change and returns the Notifications received.

        It returns the Updates before and after receiving a sync_response.

        All Updates received are accumulated up to when the channel is
         closed (by server, by timeout or error).

        Optionally, it calls a callback function when sync_response is
          received.

        Does not support heartbeat. nor OnlyUpdates (the initial update is
          needed for value change comparison).

        Args:
            xpath: gNMI path to subscribe to.
            timeout: Seconds to keep the gRPC channel open.
            sync_response_callback: Function to call upon sync_response.

        Returns:
            2 Lists of gnmi_pb2.Notification objects received,
              before and after sync_response.

        Raises:
            RpcError if unable to connect to Target.
            ValueError if the SubscribeResponse is invalid.
            GnmiError if no sync_response was received, when
                check_sync_response.
            BaseError if callback raised an Exception.
        """
        before = []
        after = []
        got_sync_response = False
        request = schema.gNMISubscriptionStreamOnChangeRequest(
            [xpath])
        try:
            for resp in self.stub.Subscribe(
                    iter([request]), timeout=timeout,
                    metadata=self._GnmiAuthMetadata()):
                if resp.sync_response:
                    got_sync_response = True
                    if sync_response_callback is not None:
                        try:
                            sync_response_callback()
                        except Exception as err:
                            raise BaseError("Callback %s Exception",
                                            sync_response_callback) from err
                elif resp.update:
                    if got_sync_response:
                        after.append(resp.update)
                    else:
                        before.append(resp.update)
                else:
                    raise ValueError("Invalid SubscribeResponse %s" % resp)
        except _MultiThreadedRendezvous as err:
            if err.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
                return before, after
            else:
                raise RpcError(err) from err
        except _InactiveRpcError as err:
            raise RpcError(err) from err

        if not got_sync_response:
            raise schema.GnmiError(
                "No Response with sync_response was received")

        return before, after



"""Test cases based on gNMI Subscribe SAMPLE requests."""

import collections

from oc_config_validate import schema, testbase

SECS_TO_NSEC = 1000000000


class SubsSampleTestCase(testbase.TestCase):
    """Subscribes SAMPLE and checks the response elements.

    Asserts that the Updates timestamps correspond to the sample interval,
        with a maximum tolerated drift.
    Asserts that all Update paths have the same number of updates.

    Args:
        xpath: gNMI path to subscribe to.
        sample_interval: Seconds interval between Updates.
        sample_timeout: Seconds to keep the Subscription and collect Updates.
        max_timestamp_drift_secs: Maximum drift for the timestamp(s) in the
            reply.
    """
    max_timestamp_drift_secs = 1
    xpath = None
    sample_interval = None
    sample_timeout = None

    def subscribeSample(self):
        """"""

        def _isTimestampDiffOk(time_diff, interval, max_drift):
            """Returns true if the timestamp differences is within limits.

            Timestamp diff must be interval +- max_drift.

            Args:
              time_diff: Timestamp diff in nsecs.
              interval: Sampling interval in nsecs.
              max_drift: Maximum drift in nsecs.

            """
            drift = time_diff - interval
            drift = drift * (-1) if drift < 0 else drift
            return (drift <= max_drift)

        self.assertArgs(["xpath", "sample_interval", "sample_timeout"])
        self.assertXpath(self.xpath)

        notifications = self.gNMISubsStreamSample(
            self.xpath,
            self.sample_interval * SECS_TO_NSEC,
            timeout=self.sample_timeout)
        self.assertIsNotNone(notifications, "No gNMI Subscribe response")

        self.responses = collections.defaultdict(list)
        for n in notifications:
            timestamp = n.timestamp
            for u in n.update:
                self.responses[schema.pathToString(u.path)].append(
                    (timestamp, schema.typedValueToPython(u.val)))

        want = (self.sample_timeout // self.sample_interval) + 1
        if self.responses:
            for path, updates in self.responses.items():
                got = len(updates)
                self.assertEqual(
                    got, want,
                    f"{got} Updates for path {path}, wanted {want}")
                if len(updates) > 1:
                    timestamp_0, _ = updates[0]
                    timeline = [0]
                    for timestamp_1, _ in updates[1:]:
                        timestamp_diff = timestamp_1 - timestamp_0
                        timeline.append(timestamp_diff)
                        self.assertTrue(
                            _isTimestampDiffOk(
                                timestamp_diff,
                                self.sample_interval * SECS_TO_NSEC,
                                self.max_timestamp_drift_secs * SECS_TO_NSEC),
                            f"Update timestamps for '{path}' out of interval: "
                            f"{timeline} diff in nsecs")
                        timestamp_0 = timestamp_1


class CountUpdates(SubsSampleTestCase):
    """Subscribes SAMPLE and checks the returned update paths.

    This tests that a Subscription consistenly reports all Update paths.

    Args:
        xpath: gNMI paths to subscribe to. Can contain wildcards.
        updates_count: Number of expected disctinct Update paths, per reply.
    """
    update_paths_count = None

    def testSubscribeSample(self):
        """"""
        self.assertArgs(["update_paths_count"])
        self.subscribeSample()
        got_paths = len(self.responses)
        self.assertEqual(
            got_paths, self.update_paths_count,
            f"Expected {self.update_paths_count} Update paths, "
            f"got: {got_paths}")


class CheckLeafs(SubsSampleTestCase):
    """Subscribes SAMPLE and checks the updates againts the state OC model.

    Tests that a Sample Subscription consistenly reports all Update paths, and
    that the paths corresponds to Leafs in the OC model.

    Args:
        xpath: gNMI paths to subscribe to. Can contain wildcards only on the
          keys.
        model: Python binding class to check the replies against.
    """
    model = None

    def testSubscribeSample(self):
        """"""
        self.assertArgs(["model", "xpath"])

        self.assertModelXpath(self.model, self.xpath)

        self.subscribeSample()

        got_paths = len(self.responses)
        self.assertGreater(got_paths, 0,
                           "There are no Update replies to the Subscription")
        if self.responses:
            for path, _ in self.responses.items():
                self.assertTrue(
                    schema.isPathInRequestedPaths(path, [self.xpath]),
                    f"Unexpected update path {path} for subscription")



"""Copyright 2023 Google LLC.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.

You may obtain a copy of the License at
                https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import unittest

from oc_config_validate import context, runner, target, testcases


class TestGetTestSuites(unittest.TestCase):
    """Test for method runner.getTestSuites()."""

    def test_getTestSuites(self):
        tgt = target.TestTarget(context.Target())
        tests = [
            context.TestCase("SetConfigCheckState",
                             "config_state.SetConfigCheckState",
                             None),
            context.TestCase("AddStaticRoute",
                             "static_route.AddStaticRoute",
                             {"arg1": "foo", "arg2": "bar"}),
        ]
        ctx = context.TestContext("example", [], [], None, tests)
        suites = runner.getTestSuites(ctx, tgt)
        self.assertIsInstance(
            suites[0]._tests[0], testcases.config_state.SetConfigCheckState)
        self.assertIsInstance(suites[1]._tests[0],
                              testcases.static_route.AddStaticRoute)
        for s in suites:
            self.assertEqual(s._tests[0].test_target, tgt)
        for t in suites[1]._tests:
            self.assertEqual(getattr(t, "arg1", None), "foo")
            self.assertEqual(getattr(t, "arg2", None), "bar")

    def test_getTestSuites_error(self):
        tgt = target.TestTarget(context.Target())
        tests = [
            context.TestCase("notAClass",
                             "config_state.notAClass",
                             None),
            context.TestCase("NotAModule",
                             "not_a_module.aClass",
                             None),
        ]
        ctx = context.TestContext("example", [], [], None, tests)
        suites = runner.getTestSuites(ctx, tgt)
        self.assertEqual(suites, [])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from operator import attrgetter
from pyangbind.lib.yangtypes import RestrictedPrecisionDecimalType
from pyangbind.lib.yangtypes import RestrictedClassType
from pyangbind.lib.yangtypes import TypedListType
from pyangbind.lib.yangtypes import YANGBool
from pyangbind.lib.yangtypes import YANGListType
from pyangbind.lib.yangtypes import YANGDynClass
from pyangbind.lib.yangtypes import ReferenceType
from pyangbind.lib.base import PybindBase
from collections import OrderedDict
from decimal import Decimal
from bitarray import bitarray
import six

# PY3 support of some PY2 keywords (needs improved)
if six.PY3:
  import builtins as __builtin__
  long = int
elif six.PY2:
  import __builtin__

class openconfig_rsvp_sr_ext(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rsvp-sr-ext - based on the path /openconfig-rsvp-sr-ext. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This module adds extensions to the OpenConfig MPLS models to
provide extensions which allow the coexistence of RSVP-TE and
Segment Routing (SR) within the same network. It augments the
existing OpenConfig segment routing (SR) and RSVP-TE models
where required.
  """
  _pyangbind_elements = {}

  

class openconfig_segment_routing_types(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-segment-routing-types - based on the path /openconfig-segment-routing-types. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Types associated with a network instance
  """
  _pyangbind_elements = {}

  

class openconfig_srte_policy(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-srte-policy - based on the path /openconfig-srte-policy. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This module defines a collection of segment routing traffic
engineering policy operational states.

Each policy, identified by a combination of color and endpoint,
has one or more candidate paths learned from one or more sources.
The best valid/usable path is marked as active and programmed in
forwarding plane.

A candidate path, identified by protocol-origin, originator and
discriminator, can have one and more segment-list defining the
path traffic should take. Each segment-list is associated with a
weight for weighted load balancing.

Traffic counters related to SR policies are also defined in this
module.
  """
  _pyangbind_elements = {}

  

class openconfig_segment_routing(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-segment-routing - based on the path /openconfig-segment-routing. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Configuration and operational state parameters relating to the
segment routing. This module defines a number of elements which are
instantiated in multiple places throughout the OpenConfig collection
of models.

Particularly:
 - SRGB+LB dataplane instances - directly instantied by SR.
 - SRGB+LB dataplane reservations - instantiated within MPLS and future SR
                                 dataplanes.
 - SR SID advertisements - instantiated within the relevant IGP.
 - SR-specific counters - instantied within the relevant dataplane.
  """
  _pyangbind_elements = {}

  




"""Test cases based on gNMI Set/Get requests."""

import json

from oc_config_validate import schema, testbase


class SetConfigCheckState(testbase.TestCase):
    """Configures on the /config container and checks the /state container.

    1. The intended JSON-IETF configuration is checked for schema validity.
    1. It is sent in a gNMI Set request,  to the /config container.
    1. The same container is fetched in a gNMI Get request and checked for
         schema validity. It is compared with the sent configuration.
    1. The /state container is fetched in a gNMI Get request and checked for
         schema validity. It is compared with the sent configuration.

    All arguments are read from the Test YAML description.

    Args:
        xpath: gNMI path to write and read, without ending /config or /state.
        json_value: JSON-IETF value to check set, get and compare.
        model: Python binding class to check the JSON reply against.
    """
    xpath = ""
    json_value = None
    model = ""

    def setUp(self):
        self.assertArgs(["xpath", "model", "json_value"])
        xpath = self.xpath.rstrip("/") + "/config"
        self.assertXpath(xpath)
        self.assertModelXpath(self.model, xpath)
        self.assertIsInstance(self.json_value, dict,
                              "The value is not a valid JSON object")
        model = schema.ocContainerFromPath(self.model, xpath)
        self.assertJsonModel(json.dumps(self.json_value), model,
                             "JSON value to Set does not match the model")

    @testbase.retryAssertionError
    def _get_check_retry(self):
        # gNMI Get on /config
        xpath = self.xpath.rstrip("/") + "/config"
        resp_val = self.gNMIGetAssertJson(xpath)
        model = schema.ocContainerFromPath(self.model, xpath)
        self.assertJsonModel(resp_val, model,
                             "Get /config does not match the model")
        got = json.loads(resp_val)
        cmp, diff = schema.intersectCmp(self.json_value, got)
        self.assertTrue(cmp, diff)

        # gNMI Get on /state
        xpath = self.xpath.rstrip("/") + "/state"
        resp = self.gNMIGet(xpath)
        self.assertIsNotNone(resp, "No gNMI GET response")
        resp_val = resp.json_ietf_val
        self.assertIsNotNone(resp_val,
                             "The gNMI GET response is not JSON IETF")
        model = schema.ocContainerFromPath(self.model, xpath)
        self.assertJsonModel(resp_val, model,
                             "Get /state does not match the model")
        got = json.loads(resp_val)
        cmp, diff = schema.intersectCmp(self.json_value, got)
        self.assertTrue(cmp, diff)

    def testSetConfigCheckState(self):
        # gNMI Set on /config
        xpath = self.xpath.rstrip("/") + "/config"
        self.assertTrue(
            self.gNMISetUpdate(xpath, self.json_value),
            "gNMI Set did not succeed.")
        self._get_check_retry()


class DeleteConfigCheckState(testbase.TestCase):
    """Deletes the xpath and checks the /config and /state container after.

    1. gNMI Get request validates that the /config container exists.
    1. gNMI Delete request removes the xpath.
    1. gNMI Get request validates that the /config container no longer exists.
    1. gNMI Get request validates that the /state container no longer exists.

    All arguments are read from the Test YAML description.

    Args:
        xpath: gNMI path to delete, without ending /config or /state.
    """
    xpath = ""

    def setUp(self):
        self.assertArgs(["xpath"])
        self.assertXpath(self.xpath)

    @testbase.retryAssertionError
    def _get_check_retry(self):
        # gNMI Get on /config to verify it is deleted
        self.assertFalse(
            self.gNMIGet(self.xpath.rstrip("/") + "/config"),
            "There is still a /config container for the xpath.")

        # gNMI Get on /state to verify it is deleted
        self.assertFalse(
            self.gNMIGet(self.xpath.rstrip("/") + "/state"),
            "There is still a /state container for the xpath.")

    def testDeleteConfigCheckState(self):

        # gNMI Get on /config to verify it is configured
        self.assertTrue(
            self.gNMIGet(self.xpath.rstrip("/") + "/config"),
            "There is no configured /config container for the xpath.")

        # gNMI Delete on the xpath
        self.assertTrue(
            self.gNMISetDelete(self.xpath.rstrip("/")),
            "gNMI Delete did not succeed.")

        # gNMI Gets to confirm deletion
        self._get_check_retry()



"""Test cases based on gNMI Subscribe ONCE requests."""

import time

from oc_config_validate import schema, testbase


class SubsOnceTestCase(testbase.TestCase):
    """Subscribes ONCE and checks the response elements.

    The arguments are optional, only checked when defined.

    Args:
        xpaths: List of gNMI paths to subscribe to.
        notifications_count: Number of expected Notifications.
        max_delay_secs: Maximum delay of the replies, measured with the
          timestamp value(s).
    """
    notifications_count = None
    max_delay_secs = None
    xpaths = None

    def subscribeOnce(self):
        """"""
        for xpath in self.xpaths:
            self.assertXpath(xpath)
        self.now = int(time.time())
        self.responses = self.gNMISubsOnce(self.xpaths)
        self.assertTrue(bool(self.responses),
                        "No gNMI Subscribe response")

        if self.max_delay_secs:
            for n in self.responses:
                timestamp_diff = (n.timestamp // 1000000000) - self.now
                self.assertTrue(
                    (self.max_delay_secs * -
                     1) < timestamp_diff < self.max_delay_secs,
                    f"Timestamp diff too long: {timestamp_diff} secs")

        if self.notifications_count:
            notifications = len(self.responses)
            self.assertEqual(
                notifications, self.notifications_count,
                f"Expected {self.notifications_count} notifications, "
                f"got {notifications}")


class CountUpdatesCheckType(SubsOnceTestCase):
    """Subscribes ONCE and checks the returned updates and their value types.

    Args:
        xpaths: List of gNMI paths to subscribe to. Can contain wildcards.
        updates_count: Number of expected Update messages.
        values_type: Python type of the values of the Updates.

    """
    values_type = None
    updates_count = None

    def test100(self):
        """"""
        self.assertArgs(["xpaths", "values_type", "updates_count"])

        self.subscribeOnce()

        updates = 0
        for n in self.responses:
            updates += len(n.update)
            for u in n.update:
                got_path = schema.pathToString(u.path)
                self.assertTrue(
                    schema.isPathInRequestedPaths(got_path, self.xpaths),
                    f"Unexpected update path {got_path} for subscription")
                self.assertTrue(
                    u.val.HasField(self.values_type),
                    f"Value of Update {schema.pathToString(u.path)} "
                    f"is not of type {self.values_type}: {u.val}")

        self.assertEqual(
            updates, self.updates_count,
            f"Expected {self.updates_count} Updates, got: {updates}")


class CheckLeafs(SubsOnceTestCase):
    """Subscribes ONCE and checks the updates againts the OC model.

    All arguments are read from the Test YAML description.

    Args:
        xpaths: List of gNMI paths to subscribe to. The paths can contain
          wildcards only in the keys.
        model: Python binding class to check the JSON reply against.
        check_missing_model_paths: If True, missing OC Model leaf paths in the
                                   Updates replies are checked.
    """
    model = None
    check_missing_model_paths = False

    def test100(self):
        """"""
        self.assertArgs(["xpaths", "model"])

        want_paths = []
        for xpath in self.xpaths:
            self.assertModelXpath(self.model, xpath)
            want_paths.extend(schema.ocLeafsPaths(self.model, xpath))

        self.subscribeOnce()

        got_updates = []
        for n in self.responses:
            got_updates.extend(n.update)

        self.assertGreater(
            len(got_updates), 0,
            "There are no Updates as reply to the Subscription")

        got_paths = []
        for u in got_updates:
            got_path = schema.pathToString(u.path)
            got_paths.append(got_path)
            self.assertTrue(
                schema.isPathInRequestedPaths(got_path, want_paths),
                f"Unexpected update path {got_path} for subscription")

        if self.check_missing_model_paths:
            for want_path in want_paths:
                self.assertIn(
                    want_path, got_paths,
                    f"Missing update path {want_path} for OC model"
                    f" {self.model}, got {got_paths}")



# -*- coding: utf-8 -*-
from operator import attrgetter
from pyangbind.lib.yangtypes import RestrictedPrecisionDecimalType
from pyangbind.lib.yangtypes import RestrictedClassType
from pyangbind.lib.yangtypes import TypedListType
from pyangbind.lib.yangtypes import YANGBool
from pyangbind.lib.yangtypes import YANGListType
from pyangbind.lib.yangtypes import YANGDynClass
from pyangbind.lib.yangtypes import ReferenceType
from pyangbind.lib.base import PybindBase
from collections import OrderedDict
from decimal import Decimal
from bitarray import bitarray
import six

# PY3 support of some PY2 keywords (needs improved)
if six.PY3:
  import builtins as __builtin__
  long = int
elif six.PY2:
  import __builtin__

class openconfig_pf_interfaces(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-pf-interfaces - based on the path /openconfig-pf-interfaces. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains groupings related to the association
between interfaces and policy forwarding rules.
  """
  _pyangbind_elements = {}

  

class openconfig_pf_forwarding_policies(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-pf-forwarding-policies - based on the path /openconfig-pf-forwarding-policies. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains configuration and operational state
relating to the definition of policy-forwarding policies.
  """
  _pyangbind_elements = {}

  

class openconfig_pf_srte(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-pf-srte - based on the path /openconfig-pf-srte. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This module defines extensions to the OpenConfig policy forwarding
module to support static segment routing traffic engineering policy
definitions. Extensions are provided to match:

 - Ingress binding SIDs, such that traffic can be mapped based on
   an ingress label.
 - A colour community and endpoint combination, such that the
   routes can be resolved according to the policy forwarding
   entries that are to be installed.

In addition, policy forwarding actions associated with next-hops are
added to the model. The next-hop set to be forwarded to is augmented
to cover a set of lists of segments. The most common application of
such segment lists is to express stacks of MPLS labels which are used
as SR segments. In addition, they may be used to expressed segments
in the form of IPv6 addresses.
  """
  _pyangbind_elements = {}

  

class openconfig_pf_forwarding_policies(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-pf-forwarding-policies - based on the path /openconfig-pf-forwarding-policies. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains configuration and operational state
relating to the definition of policy-forwarding policies.
  """
  _pyangbind_elements = {}

  

class openconfig_pf_path_groups(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-pf-path-groups - based on the path /openconfig-pf-path-groups. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains configuration and operational state
relating to path-selection-groups which are used to group
forwarding entities together to be used as policy forwarding
targets.
  """
  _pyangbind_elements = {}

  

class openconfig_pf_path_groups(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-pf-path-groups - based on the path /openconfig-pf-path-groups. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains configuration and operational state
relating to path-selection-groups which are used to group
forwarding entities together to be used as policy forwarding
targets.
  """
  _pyangbind_elements = {}

  

class openconfig_policy_forwarding(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-policy-forwarding - based on the path /openconfig-policy-forwarding. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This module defines configuration and operational state data
relating to policy-based forwarding. Policy-based forwarding is
utilised when a system chooses how to forward packets (including
applying data-plane operations such as encapsulation or
decapsulation) based on policies other than destination L2 or L3
header. Typically, systems may implement:

- IP policy-based routing, where routing may be done based on the
  source plus destination of an IP packet; information within the
  L4 header; or some combination of both.
- Encapsulation or decapsulation based on certain policy
  information - for example, matching particular IP destinations
  and decapsulating GRE headers.
- Class-based selection of egress routes - such as class-based
  selection of an egress MPLS path.

The policies that are defined in this model are applied to a
particular ingress context of a network element (e.g., interface)
and are defined to apply following other interface policy such as
QoS classification and access control lists.

This module defines:

- policy-forwarding
|
|--- policies
|    |-- policy
|        |-- [match criteria]    How packets are defined to
|        |                       match policy.
|        |-- [forwarding-action] How packets matching should
|                                 be forwarded.
|--- interfaces
|    |-- interfaces
|        | -- apply-forwarding-policy  Forwarding policy to
|                                      used on the interface.
|--- path-selection-groups
    |-- path-selection-group     A group of forwarding resources
                                 that are grouped for purposes
                                 of next-hop selection.

A forwarding-policy specifies the match criteria that it intends
to use to determine the packets that it reroutes - this may
consist of a number of criteria, such as DSCP. The action of the
policy results in a forwarding action being applied to matching
packets. For example, decapsulating the packet from a GRE header.
In order to enact the policy based on particular interfaces - the
forwarding-policy is applied to an interface via referencing it
within an 'apply-forwarding-policy' statement associated with an
interface.

In some cases (e.g., Class-Based Tunnel Selection) the forwarding
action does not resolve to a single egress action, and rather
normal forwarding rules are to be applied but considering a subset
of forwarding resources. In these cases, a path-selection-group
can be created, referencing the subset of forwarding paths that
should be used for the egress selection. In the case that a subset
of MPLS LSPs are eligible for, say, DSCP 46 marked packets, a
path-selection-group is created, referencing the subset of LSPs.
The forwarding action of the corresponding policy is set to
PATH_GROUP and references the configured group of LSPs.
  """
  _pyangbind_elements = {}

  

class openconfig_pf_forwarding_policies(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-pf-forwarding-policies - based on the path /openconfig-pf-forwarding-policies. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains configuration and operational state
relating to the definition of policy-forwarding policies.
  """
  _pyangbind_elements = {}

  

class openconfig_pf_path_groups(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-pf-path-groups - based on the path /openconfig-pf-path-groups. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains configuration and operational state
relating to path-selection-groups which are used to group
forwarding entities together to be used as policy forwarding
targets.
  """
  _pyangbind_elements = {}

  

class openconfig_pf_interfaces(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-pf-interfaces - based on the path /openconfig-pf-interfaces. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains groupings related to the association
between interfaces and policy forwarding rules.
  """
  _pyangbind_elements = {}

  




"""Test cases based on gNMI Set/Get requests."""

import json

from oc_config_validate import schema, testbase


class TestCase(testbase.TestCase):
    """Sends gNMI Set and later Get.

    Args:
        xpath: gNMI path to write and read.
        json_value: JSON-IETF value to check, set and get.
        model: Python binding class to check the JSON reply against.
    """
    xpath = None
    json_value = None
    model = None

    def setUp(self):
        # Check testcase arguments
        self.assertArgs(["xpath", "json_value", "model"])
        self.assertXpath(self.xpath)
        self.assertIsInstance(self.json_value, dict,
                              "The value is not a valid JSON object")
        self.assertModelXpath(self.model, self.xpath)
        model = schema.ocContainerFromPath(self.model, self.xpath)
        self.assertJsonModel(json.dumps(self.json_value), model,
                             "JSON value to Set does not match the model")
        # Set Update
        self.assertTrue(self.gNMISetUpdate(self.xpath, self.json_value),
                        "gNMI Set did not succeed.")


class SetGetJsonCheck(TestCase):
    """Sends gNMI Set and later Get, schema-checking the JSON-IETF value.

    1. The intended JSON-IETF configuration is checked for schema validity.
    1. It is sent in a gNMI Set request.
    1. The same path is used for a gNMI Get request.
    1. The returned value is checked for schema validity

    All arguments are read from the Test YAML description.

    Args:
        xpath: gNMI path to write and read.
        json_value: JSON-IETF value to check, set and get.
        model: Python binding class to check the JSON reply against.
    """

    @testbase.retryAssertionError
    def testSetGetJsonCheck(self):
        """"""
        self.resp_val = self.gNMIGetAssertJson(self.xpath)
        model = schema.ocContainerFromPath(self.model, self.xpath)
        self.assertJsonModel(self.resp_val, model,
                             "Get response JSON does not match the model")


class SetGetJsonCheckCompare(TestCase):
    """Does what SetGetJsonCheck does, but also compares the JSON Get reply.

    1. The intended JSON-IETF configuration is checked for schema validity.
    1. It is sent in a gNMI Set request.
    1. The same path is used for a gNMI Get request.
    1. The returned value is checked for schema validity
    1. The returned value is compared with the sent value

    All arguments are read from the Test YAML description.

    Args:
        xpath: gNMI path to write and read.
        json_value: JSON-IETF value to check set, get and compare.
        model: Python binding class to check the JSON reply against.
    """

    @testbase.retryAssertionError
    def testSetGetJsonCheckCompare(self):
        self.resp_val = self.gNMIGetAssertJson(self.xpath)
        model = schema.ocContainerFromPath(self.model, self.xpath)
        self.assertJsonModel(self.resp_val, model,
                             "Get response JSON does not match the model")
        got = json.loads(self.resp_val)
        cmp, diff = schema.intersectCmp(self.json_value, got)
        self.assertTrue(cmp, diff)



"""Copyright 2021 Google LLC.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.

You may obtain a copy of the License at
                https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import importlib
import json
import re
from inspect import isclass
from typing import Any, List, Tuple, Union

from pyangbind.lib import yangtypes
from pyangbind.lib.base import PybindBase
from pyangbind.lib.serialise import pybindJSONDecoder

from oc_config_validate import models
from oc_config_validate.gnmi import gnmi_pb2  # type: ignore

_RE_GNMI_PATH_ELEM = re.compile(r'''^
(?P<name>[^[]+)   # gNMI path name
(\[(?P<key>\w\D+) # gNMI path key
=
(?P<value>.*)     # gNMI path value
\])?$
''', re.VERBOSE)

_RE_GNMI_XPATH = re.compile(r'''/(?=(?:[^\[\]]|\[[^\[\]]+\])*$)''')

_RE_GNMI_KEY = re.compile(r'\[([^]]*)\]')

type_translate_map = {int: 'int_val', float: 'float_val', bool: 'bool_val',
                      str: 'string_val', dict: 'json_ietf_val'}
value_translate_map = {'int_val': int, 'float_val': float, 'bool_val': bool,
                       'string_val': str, 'json_ietf_val': dict,
                       'json_val': dict, 'uint_val': int, 'leaflist_val': list}


class BaseError(Exception):
    """Base Exception raised by this module."""


class ModelClassError(BaseError):
    """Error parsing or loading an OC Model Class."""


class XpathError(BaseError):
    """Error parsing gNMI xpath."""


class GnmiError(BaseError):
    """Error using gNMI."""


def decodeJson(json_text: Union[str, bytes], obj: PybindBase):
    """Decode a JSON text into a PybindBase object.

    This method is used to validate the JSON text adheres to the OC schema.

    Args:
        json_text: The JSON-IETF text to decode.
        obj: The PybindBase object to decode the texto into.

    Raises:
        Error if unable to parse the JSON text.
    """
    json_text = removeOpenConfigPrefix(json_text)
    pybindJSONDecoder.load_ietf_json(json.loads(json_text), None, None, obj)


def removeOpenConfigPrefix(json_text: Union[str, bytes]) -> Union[str, bytes]:
    """Remove open-config prefixed so JSON text can be processed by PyBind.

    When JSON-IETF is used for a gNMI response, type references will
    prepend the corresponding model name(ie. openconfig-aaa:RADIUS) when
    referencing a server of type RADIUS in a leaf.  These must be removed
    before being processed by PyBind.

    Args:
        json_text: The JSON-IETF text to correct.

    Returns:
        The string that can be fed directly to pybindJSONDecoder.
    """
    # https://regex101.com/r/xiZj4Q/1
    if isinstance(json_text, bytes):
        return re.sub(b'(openconfig(-[a-z]+)+\:)', b'',
                      json_text)  # noqa
    return re.sub(r'(openconfig(-[a-z]+)+\:)', '', json_text)


def ocContainerFromPath(model: str, xpath: str) -> PybindBase:
    """Create an empty PybindBase instance of the model for the path.

    This method look for the model class in the oc_config_validate.models
      package

    Args:
        model: the OC model class name in the oc_config_validate.models
        package, as `module.class`.
        xpath: the xpath to the OC container to create.

    Returns:
        An PybindBase object of the class.

    Raises:
        ModelClassError if unable to find the Python class or if the class is
          not derived from PybindBase.
        XpathError if the xpath is invalid.
        AttributeError if unable to find an xpath element in the OC class.
    """
    model_inst = _ocObjFromPath(model, xpath)
    if not issubclass(model_inst.__class__, PybindBase):
        raise ModelClassError(
            "%s:%s is not a valid container class" % (model, xpath))
    return model_inst


def ocLeafFromPath(model: str, xpath: str) -> Any:
    """Create an empty value of the model leaf pointed by the xpath.

    This method look for the model type in the oc_config_validate.models
      package

    Args:
        model: the OC model class name in the oc_config_validate.models
          package, as `module.class`.
        xpath: the xpath to the OC leaf to create.

    Returns:
        A value object of the leaf, like int, float, str, etc.

    Raises:
        ModelClassError if unable to find the Python class or if the instance
          is not a Leaf.
        XpathError if the xpath is invalid.
        AttributeError if unable to find an xpath element in the OC class.
    """
    model_inst = _ocObjFromPath(model, xpath)
    if issubclass(model_inst.__class__, PybindBase):
        raise ModelClassError(
            "%s:%s is not a valid leaf class" % (model, xpath))
    return model_inst


def _ocObjFromPath(model: str, xpath: str) -> Any:
    """Create an empty Object of the model for the path.

    This method look for the model class in the oc_config_validate.models
      package

    Args:
        model: the OC model class name in the oc_config_validate.models
        package, as `module.class`.
        xpath: the xpath to the OC container to create.

    Returns:
        An object of the class.

    Raises:
        ModelClassError if unable to find the Python class or if the class.
        XpathError if the xpath is invalid.
        AttributeError if unable to find an xpath element in the OC class.
    """

    parts = model.split('.')
    if len(parts) != 2:
        raise ModelClassError("%s is not module.class" % model)
    model_mod = importlib.import_module(
        "oc_config_validate.models." + parts[0])
    model_cls = getattr(model_mod, parts[1])
    if not isclass(model_cls):
        raise ModelClassError(
            "%s is not a class in oc_config_validate.models package" % model)

    gnmi_xpath = parsePath(xpath)

    model_inst = model_cls()
    for e in gnmi_xpath.elem:
        model_inst = getattr(model_inst, yangtypes.safe_name(e.name))
        if e.key:
            save_key = {}
            for k, v in e.key.items():
                save_key[yangtypes.safe_name(k)] = v
            model_inst = model_inst.add(**save_key)
    return model_inst


def fixSubifIndex(json_value: dict):
    """Rewrite the index of a pybindJSON-produced subinterface as int.

    pybindJSON dumps the index as a str value, instead of int.

    https://github.com/robshakir/pyangbind/issues/139

    """
    index = json_value['openconfig-interfaces:subinterfaces'][
        'subinterface'][0]['index']
    json_value['openconfig-interfaces:subinterfaces']['subinterface'][0][
        'index'] = int(index)


def getOcModelsVersions() -> str:
    """Returns a multi-line string of the OC models versions used."""
    return models.__model_versions__.strip()


def ocLeafsPaths(model: str, xpath: str) -> List[str]:
    """Returns a list of xpaths to leafs, rooted at xpath.

    This method look for the model class in the oc_config_validate.models
      package

    Args:
        model: the OC model class name in the oc_config_validate.models
        package, as `module.class`.
        xpath: the xpath to the OC container to root at.

    Returns:
        A list of paths to leafs, rooted at the xpath.

    Raises:
        Error if unable to find the Python class or if the class is not derived
          from PybindBase.
        XpathError if the xpath is invalid.
        AttributeError if unable to find an xpath element in the OC class.

    """
    obj = ocContainerFromPath(model, xpath)

    paths = []
    for leaf, val in obj.get().items():
        if type(val) == dict:
            paths.extend(ocLeafsPaths(model, xpath + "/" + leaf))
        else:
            paths.append(xpath + "/" + leaf)
    return paths


def parsePath(xpath: str) -> gnmi_pb2.Path:
    """Parse an xpath string into gNMI.Path object.

    Args:
        xpath: String representation of a gNMI path.

    Returs:
        A gNMI.Path oject.

    Raises:
        XpathError: If unable to parse the xpath provided.
    """

    if not xpath or xpath == '/':
        return gnmi_pb2.Path(elem=[])
    xpath = xpath.strip('/').strip('/')  # Removes leading/trailing '/'.
    path_parts = _RE_GNMI_XPATH.split(xpath)
    gnmi_elems = []
    for part in path_parts:
        part_search = _RE_GNMI_PATH_ELEM.search(part)
        if not part_search:  # Invalid path specified.
            raise XpathError('xpath component parse error: %s' % part)
        if part_search.group('key') is not None:  # A path key was provided.
            key = {}
            for k in _RE_GNMI_KEY.findall(part):
                key[k.split('=')[0]] = k.split('=')[-1]
            gnmi_elems.append(
                gnmi_pb2.PathElem(name=part_search.group('name'), key=key))
        else:
            gnmi_elems.append(gnmi_pb2.PathElem(name=part, key={}))
    return gnmi_pb2.Path(elem=gnmi_elems)


def isPathOK(xpath: str) -> bool:
    """Return True is the xpath is parsed correctly into a gNMI Path

    Args:
        xpath: The xpath to parse
    """
    try:
        parsePath(xpath)
    except XpathError:
        return False
    return True


def isPathInRequestedPaths(xpath: str, xpaths: List[str]) -> bool:
    """Returns True if xpath in included or equal to any request path.

    The request paths can contain paths with wildcard '*'.

    Args:
      xpath: Non-wildcard xpath to check.
      xpaths: List of xpaths (can have wildcards) to contain xpath.
    """
    def _inOrEqual(x: str):
        # Ensure the path has keys sorted
        x = pathToString(parsePath(x))
        x_regex = re.escape(x).replace(r"\*", r"(.+)")
        return re.match(x_regex, xpath) is not None

    xpath = pathToString(parsePath(xpath))
    return any(map(_inOrEqual, xpaths))


def pathToString(path: gnmi_pb2.Path) -> str:
    """Parse a gNMI Path to a URI-like string.

    The keys are sorted alphabetically.

    Args:
        A gNMI.Path oject.

    Returs:
        xpath: String representation of a gNMI path.

    Raises:
        XpathError: If unable to parse the xpath provided.
    """
    path_elems = []
    for e in path.elem:
        elem = e.name
        if hasattr(e, "key"):
            keys = [f"[{k}={v}]" for k, v in e.key.items()]
            elem += f"{''.join(sorted(keys))}"
        path_elems.append(elem)
    return "/" + "/".join(path_elems)


def notificationsJsonString(notifications: List[gnmi_pb2.Notification],
                            indent: int = 2) -> str:
    """Returns a JSON-idented string of the Notification messages.

    Args:
        notifications: List of Notification messages to format.
        indent: number of JSON indentation spaces
    """
    notifs = []
    for n in notifications:
        notif = {"timestamp": n.timestamp, "updates": [None]*len(n.update)}
        for i, u in enumerate(n.update):
            notif["updates"][i] = {"path": pathToString(
                u.path), "value": typedValueToPython(u.val)}
        notifs.append(notif)
    return json.dumps(notifs, indent=2, default=str)


def pythonToTypedValue(value: Any) -> gnmi_pb2.TypedValue:
    """Return the gNMI TypedValue for a Python variable.

    Args:
        value: Value to set.

    Raises:
        GnmiError if type is unsupported.
    """

    typed_val = gnmi_pb2.TypedValue()
    try:
        value_type = type_translate_map[type(value)]
    except KeyError:
        raise GnmiError(
            f"Value type '{type(value)}' cannot be made a gNMI TypedValue")

    if isinstance(value, dict):
        setattr(typed_val, value_type, json.dumps(value).encode())
    else:
        setattr(typed_val, value_type, value)
    return typed_val


def typedValueToPython(value: gnmi_pb2.TypedValue) -> Union[
        bool, int, float, dict, str, list]:
    """Return the Python variable for a gNMI TypedValue.

    Args:
        value: gNMI TypedValue to convert.
    """
    value_type = value.WhichOneof("value")
    if value_type in ["json_ietf_val", "json_val"]:
        return json.loads(getattr(value, value_type))
    if value_type == "leaflist_val":
        return [typedValueToPython(v) for v in value.leaflist_val.element]
    else:
        tp = value_translate_map[value_type]
        return tp(getattr(value, value_type))
    raise GnmiError(f"Unsupported Type {value_type}")


def intersectCmp(want: dict, got: dict) -> Tuple[bool, str]:
    """Return true if all the keys of want are the same in got.

    E.g.: Taking want={"foo": 0, "bla": "test"} and
        got={"foo": 1, "bar": true, "bla": "test"}, the keys "foo"
        and "bla" are looked for and compared in got.

    The same comparison applies to nested dictionaries and lists.

    Args:
        want, got: Dictionaries to compare.

    Returns:
        A tuple with the comparison result and a diff text, if different.
    """
    for k in want.keys():
        x = want[k]
        y = got.get(k)
        if y is None:
            return False, "key %s not found" % k
        if isinstance(x, dict) and isinstance(y, dict):
            return intersectCmp(x, y)
        if isinstance(x, list) and isinstance(y, list):
            return intersectListCmp(x, y)
        if x != y:
            return False, "key %s: got '%s', wanted '%s'" % (k, y, x)
    return True, ""


def intersectListCmp(want: list, got: list) -> Tuple[bool, str]:
    """Return true if all the elements in want are the same in got.

    For dictionaries, intersectCmp() is used for comparison.
    Order is not considered.

    E.g.: Taking [0, "bla"] and [1, "bar", "bla"], the elements 0 and "bla"
        are looked for and compared in got.

    The same comparison applies to nested dictionaries and lists.

    Args:
        want, got: Lists to compare.

    Returns:
        A tuple with the comparison result and a diff text, if different.
    """
    for i in want:
        for j in got:
            matched = False
            if type(i) != type(j):
                continue
            if isinstance(i, dict):
                cmp, _ = intersectCmp(i, j)
            else:
                cmp = bool(i == j)
            if cmp:
                matched = True
                break
        if not matched:
            return False, "List element %s not matched" % i
    return True, ""


def gNMISubscriptionOnceRequest(
    xpaths: List[str],
    encoding: str = 'JSON_IETF'
) -> gnmi_pb2.SubscribeRequest:
    """
    Returns a gNMI Subscription ONCE requests for the xpaths.

    Args:
        xpaths: List of gNMI xpaths to subscribe to.
        encoding: Encoding requested for the Updates.
    """
    paths = [parsePath(xpath) for xpath in xpaths]
    return gnmi_pb2.SubscribeRequest(subscribe=gnmi_pb2.SubscriptionList(
        subscription=[gnmi_pb2.Subscription(path=path)
                      for path in paths],
        mode=gnmi_pb2.SubscriptionList.ONCE,
        encoding=encoding))


def gNMISubscriptionStreamSampleRequest(
    xpaths: List[str],
    sample_interval: int,
    encoding: str = 'JSON_IETF'
) -> gnmi_pb2.SubscribeRequest:
    """
    Returns a gNMI Subscription STREAM SAMPLE request for the xpaths.

    Args:
        xpaths: gNMI xpaths to subscribe to.
        sample_interval: Nanoseconds interval between Updates.
        encoding: Encoding requested for the Updates.
    """
    paths = [parsePath(xpath) for xpath in xpaths]
    return gnmi_pb2.SubscribeRequest(subscribe=gnmi_pb2.SubscriptionList(
        subscription=[gnmi_pb2.Subscription(
            path=path,
            mode=gnmi_pb2.SubscriptionMode.SAMPLE,
            sample_interval=sample_interval) for path in paths],
        mode=gnmi_pb2.SubscriptionList.STREAM,
        encoding=encoding))


def gNMISubscriptionStreamOnChangeRequest(
    xpaths: List[str],
    encoding: str = 'JSON_IETF'
) -> gnmi_pb2.SubscribeRequest:
    """
    Returns a gNMI Subscription STREAM ON_CHANGE request for the xpaths.

    Args:
        xpaths: gNMI xpaths to subscribe to.
        encoding: Encoding requested for the Updates.
    """
    paths = [parsePath(xpath) for xpath in xpaths]
    return gnmi_pb2.SubscribeRequest(subscribe=gnmi_pb2.SubscriptionList(
        subscription=[gnmi_pb2.Subscription(
            path=path,
            mode=gnmi_pb2.SubscriptionMode.ON_CHANGE) for path in paths],
        mode=gnmi_pb2.SubscriptionList.STREAM,
        encoding=encoding))



"""Copyright 2021 Google LLC.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.

You may obtain a copy of the License at
                https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import json

from oc_config_validate import testbase

SUPPORTED_FORMATS = ["json"]


class FormatterError(Exception):
    """Error while working with Formatter"""


class Formatter():
    """Interface for concrete results formatters."""

    def writeResultsToFile(self, test_run: testbase.TestRun, filename: str):
        """Templates writing to a file."""
        raise FormatterError("writeResultsToFile() not implemented.")


class JSONFormatter(Formatter):
    """Writes the test results to a JSON file"""

    def writeResultsToFile(self, test_run: testbase.TestRun, filename: str):
        """Write the results of a test run to a JSON file.

        This  method assumes the filename is valid and writable.

        """
        with open(filename, 'w+', encoding='utf-8') as file:
            json.dump(test_run, file, indent=2, default=lambda o: o.__dict__)


def makeFormatter(format_name: str) -> Formatter:
    """Return a Formatter for the format, if supported.

    Args:
        format_name: a valid format from SUPPORTED_FORMATS.

    Raises:
        FormatterError if the format is not supported.
    """
    if format_name.lower() == "json":
        return JSONFormatter()
    raise FormatterError("%s is not a supported format." % format_name)



__model_versions__ = """openconfig-icmpv6-types@2023-05-02
openconfig-packet-match@2023-03-01
openconfig-packet-match-types@2023-01-29
openconfig-icmpv4-types@2023-01-26
openconfig-acl@2023-02-06
openconfig-aft-types@2022-05-05
openconfig-aft-pf@2023-04-19 (belongs-to openconfig-aft)
openconfig-aft@2023-04-19
openconfig-aft-ipv6@2023-04-19 (belongs-to openconfig-aft)
openconfig-aft-common@2023-04-19 (belongs-to openconfig-aft)
openconfig-aft-ethernet@2023-04-19 (belongs-to openconfig-aft)
openconfig-aft-ipv4@2023-04-19 (belongs-to openconfig-aft)
openconfig-aft-mpls@2023-04-19 (belongs-to openconfig-aft)
openconfig-aft-state-synced@2023-04-19 (belongs-to openconfig-aft)
openconfig-aft-network-instance@2023-04-25
openconfig-ate-intf@2019-08-07
openconfig-ate-flow@2022-04-29
openconfig-bfd@2023-02-06
openconfig-bgp-common-multiprotocol@2023-06-27 (belongs-to openconfig-bgp)
openconfig-bgp-global@2023-06-27 (belongs-to openconfig-bgp)
openconfig-bgp-peer-group@2023-06-27 (belongs-to openconfig-bgp)
openconfig-bgp@2023-06-27
openconfig-bgp-policy@2023-03-27
openconfig-bgp-common-structure@2023-06-27 (belongs-to openconfig-bgp)
openconfig-bgp-neighbor@2023-06-27 (belongs-to openconfig-bgp)
openconfig-bgp-errors@2023-03-31 (belongs-to openconfig-bgp-types)
openconfig-bgp-common@2023-06-27 (belongs-to openconfig-bgp)
openconfig-bgp-types@2023-03-31
openconfig-catalog-types@2018-11-21
openconfig-module-catalog@2018-11-21
openconfig-defined-sets@2022-12-14
openconfig-terminal-device-properties@2022-04-26
openconfig-terminal-device-property-types@2022-03-08
openconfig-codegen-extensions@2020-10-05
openconfig-metadata@2020-08-06
openconfig-fw-high-availability@2021-06-16
openconfig-fw-link-monitoring@2021-06-16
openconfig-gribi@2021-04-06
openconfig-if-8021x@2020-01-28
openconfig-if-ip-ext@2018-11-21
openconfig-if-tunnel@2018-11-21
openconfig-if-ethernet@2023-03-10
openconfig-if-ip@2023-04-12
openconfig-if-sdn-ext@2021-03-30
openconfig-if-aggregate@2022-06-28
openconfig-if-ethernet-ext@2018-11-21
openconfig-if-poe@2018-11-21
openconfig-interfaces@2023-02-06
openconfig-isis-types@2022-02-11
openconfig-isis-routing@2023-05-01 (belongs-to openconfig-isis)
openconfig-isis-policy@2023-04-28
openconfig-isis-lsp@2023-05-01 (belongs-to openconfig-isis)
openconfig-isis-lsdb-types@2018-11-21
openconfig-isis@2023-05-01
openconfig-keychain@2022-11-05
openconfig-keychain-types@2022-03-01
openconfig-lacp@2021-07-20
openconfig-lldp@2018-11-21
openconfig-lldp-types@2018-11-21
openconfig-local-routing@2022-11-01
openconfig-macsec@2023-06-08
openconfig-macsec-types@2019-07-01
openconfig-mpls-sr@2018-11-21
openconfig-mpls-ldp@2023-02-06
openconfig-mpls-rsvp@2023-02-06
openconfig-mpls-types@2021-12-01
openconfig-mpls-te@2023-04-28 (belongs-to openconfig-mpls)
openconfig-mpls-igp@2023-04-28 (belongs-to openconfig-mpls)
openconfig-mpls@2023-04-28
openconfig-mpls-static@2023-04-28 (belongs-to openconfig-mpls)
openconfig-igmp@2023-02-03
openconfig-pim@2023-03-06
openconfig-igmp-types@2018-11-21
openconfig-pim-types@2018-11-21
openconfig-network-instance-l3@2022-11-08
openconfig-route-summary@2022-12-21
openconfig-network-instance-policy@2018-11-21
openconfig-network-instance@2023-04-25
openconfig-programming-errors@2022-10-11
openconfig-network-instance-l2@2023-04-25 (belongs-to openconfig-network-instance)
openconfig-evpn@2023-03-08
openconfig-evpn-types@2021-06-21
openconfig-network-instance-types@2021-07-14
openconfig-openflow-types@2022-05-24
openconfig-openflow@2018-11-21
openconfig-terminal-device@2021-07-29
openconfig-transport-line-connectivity@2019-06-27
openconfig-optical-attenuator@2019-07-19
openconfig-wavelength-router@2021-07-26
openconfig-transport-line-common@2019-06-03
openconfig-transport-types@2023-02-08
openconfig-channel-monitor@2019-10-24
openconfig-optical-amplifier@2019-12-06
openconfig-transport-line-protection@2018-11-21
openconfig-ospfv2-lsdb@2023-03-06 (belongs-to openconfig-ospfv2)
openconfig-ospf-policy@2018-11-21
openconfig-ospf-types@2018-11-21
openconfig-ospfv2-common@2023-03-06 (belongs-to openconfig-ospfv2)
openconfig-ospfv2@2023-03-06
openconfig-ospfv2-global@2023-03-06 (belongs-to openconfig-ospfv2)
openconfig-ospfv2-area@2023-03-06 (belongs-to openconfig-ospfv2)
openconfig-ospfv2-area-interface@2023-03-06 (belongs-to openconfig-ospfv2)
openconfig-p4rt@2022-08-19
openconfig-pcep@2023-04-25
openconfig-platform-healthz@2023-04-11
openconfig-platform-common@2023-02-13 (belongs-to openconfig-platform)
openconfig-platform-fan@2018-11-21
openconfig-platform-psu@2018-11-21
openconfig-platform-types@2023-06-27
openconfig-platform-ext@2018-11-21
openconfig-platform-fabric@2022-07-28
openconfig-platform-software@2021-06-16
openconfig-platform-port@2023-03-22
openconfig-platform-transceiver@2023-05-03
openconfig-platform-linecard@2023-02-13
openconfig-platform-controller-card@2022-07-28
openconfig-platform-integrated-circuit@2022-04-20
openconfig-platform-pipeline-counters@2023-02-03
openconfig-platform-cpu@2018-11-21
openconfig-platform@2023-02-13
openconfig-pf-interfaces@2023-04-25 (belongs-to openconfig-policy-forwarding)
openconfig-pf-srte@2019-10-15
openconfig-pf-forwarding-policies@2023-04-25 (belongs-to openconfig-policy-forwarding)
openconfig-pf-path-groups@2023-04-25 (belongs-to openconfig-policy-forwarding)
openconfig-policy-forwarding@2023-04-25
openconfig-policy-types@2022-11-08
openconfig-routing-policy@2022-05-24
openconfig-probes@2018-11-21
openconfig-probes-types@2018-11-21
openconfig-qos-elements@2023-04-25 (belongs-to openconfig-qos)
openconfig-qos-mem-mgmt@2023-04-25 (belongs-to openconfig-qos)
openconfig-qos-interfaces@2023-04-25 (belongs-to openconfig-qos)
openconfig-qos@2023-04-25
openconfig-qos-types@2018-11-21
openconfig-relay-agent@2023-02-06
openconfig-rib-bgp-ext@2019-04-25
openconfig-rib-bgp-tables@2022-12-20 (belongs-to openconfig-rib-bgp)
openconfig-rib-bgp@2022-12-20
openconfig-rib-bgp-attributes@2022-12-20 (belongs-to openconfig-rib-bgp)
openconfig-rib-bgp-shared-attributes@2022-12-20 (belongs-to openconfig-rib-bgp)
openconfig-rib-bgp-table-attributes@2022-12-20 (belongs-to openconfig-rib-bgp)
openconfig-rib-bgp-types@2019-03-14
openconfig-sampling-sflow@2022-06-21
openconfig-sampling@2022-06-21
openconfig-rsvp-sr-ext@2019-07-09
openconfig-segment-routing-types@2020-02-04
openconfig-srte-policy@2021-07-28
openconfig-segment-routing@2023-02-06
openconfig-spanning-tree-types@2021-06-16
openconfig-spanning-tree@2019-11-28
openconfig-system-utilization@2023-02-13
openconfig-system-controlplane@2023-03-03
openconfig-aaa-types@2018-11-21
openconfig-system-terminal@2018-11-21
openconfig-aaa@2022-07-29
openconfig-license@2020-04-22
openconfig-alarm-types@2018-11-21
openconfig-system-grpc@2022-04-19
openconfig-aaa-tacacs@2022-07-29 (belongs-to openconfig-aaa)
openconfig-alarms@2019-07-09
openconfig-procmon@2019-03-15
openconfig-system@2023-06-16
openconfig-messages@2018-08-13
openconfig-system-logging@2023-05-04
openconfig-aaa-radius@2022-07-29 (belongs-to openconfig-aaa)
openconfig-system-bootz@2023-06-16
openconfig-telemetry@2018-11-21
openconfig-telemetry-types@2018-11-21
openconfig-yang-types@2021-07-14
openconfig-inet-types@2023-02-06
openconfig-types@2019-04-16
openconfig-vlan@2023-02-07
openconfig-vlan-types@2022-05-24
openconfig-ap-manager@2023-05-26
openconfig-access-points@2023-06-26
openconfig-wifi-mac@2023-05-26
openconfig-ap-interfaces@2023-06-26
openconfig-wifi-phy@2023-05-26
openconfig-wifi-types@2022-05-26
"""


"""Copyright 2021 Google LLC.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.

You may obtain a copy of the License at
                https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import io
import json
import logging
import threading
import time
import unittest
from typing import List, Optional

from oc_config_validate import context, schema, target, testbase, testcases


class BaseError(Exception):
    """Base Error for the runner class"""


class InitConfigError(BaseError):
    """Error applying initial configs."""


class TestClassError(BaseError):
    """Error rendering a testclass."""


class TestThread(threading.Thread):
    """Threaded class that runs tests."""

    def __init__(self, test_suite: testbase.TestSuite,
                 connection_lock: threading.Lock,
                 stop_on_error: bool = False):
        """Initialize the threaded class.

        Args:
            test_suite: A TestSuite to run in this Thread.
            node_locks: A mutex to control the use of the connection.
            stop_on_error: True if the run stops after the first error.
        """
        super().__init__()
        self.test_suite = test_suite
        self.conn_lock = connection_lock
        self.results = testbase.TestResult()
        self.stream = io.StringIO()
        self.duration_sec = 0
        self.stop_on_error = stop_on_error

    def run(self):
        """Run the test suite."""
        text_test_runner = getRunner(self.stream)
        text_test_runner.failfast = self.stop_on_error
        if self.conn_lock:
            self.conn_lock.acquire()
        start_time = int(time.time())
        self.results = text_test_runner.run(self.test_suite)
        self.duration_sec = int(time.time() - start_time)
        if self.conn_lock:
            self.conn_lock.release()
        self.results.test_class_name = self.test_suite.test_class_name
        self.results.test_name = self.test_suite.test_name
        self.results.duration_sec = self.duration_sec


def getRunner(stream: io.StringIO) -> unittest.TextTestRunner:
    """Generate a unittest text runner that uses the the StringIO stream.

    Args:
        stream: A io.StringIO object.

    Returns:
        A unittest.TextTestRunner object.
    """
    return unittest.TextTestRunner(
        stream=stream,
        buffer=False,
        verbosity=2,
        resultclass=testbase.TestResult)


def getTestSuites(ctx: context.TestContext,
                  tgt: target.TestTarget):
    suites = []
    for t in ctx.tests:
        s = makeTestSuite(t.class_name, t.name)
        if s is not None:
            s.insertTarget(tgt)
            if hasattr(t, 'args') and t.args:
                s.insertArgs(t.args)
            suites.append(s)
    return suites


def makeTestSuite(test_class_name: str,
                  test_name: str) -> Optional[testbase.TestSuite]:
    """Build a test suite of the provided test class.

    Args:
        test_class_name: The class name.
        test_name: The name of the test.

    Returns:
        A unittest.TestSuite object,,
        None if the test_class_name was not found.
    """

    def strCmp(a, b):
        """Comparer method to get alphabetically-ordered strings."""
        return (a > b) - (a < b)

    try:
        test_class = _getClassFromName(test_class_name)
    except TestClassError as err:
        logging.error(err)
        return None
    suite = unittest.makeSuite(test_class,  # pytype: disable=wrong-arg-types
                               sortUsing=strCmp,
                               prefix='test',
                               suiteClass=testbase.TestSuite)
    suite.test_class_name = test_class_name
    suite.test_name = test_name
    return suite    # pytype: disable=bad-return-type


def _getClassFromName(class_name: str) -> testbase.TestCase:
    """Find a class type from a string.

    This method interprets the class_name as module.class

    Args:
        class_name: a string with the class name

    Returns:
        A class type derived from test_testbase.TestCase.

    Raises:
        TestClassError if not found or not derived from test_testbase.TestCase.

    """
    try:
        cls = testcases
        for part in class_name.split('.'):
            cls = getattr(cls, part)
    except AttributeError as err:
        raise TestClassError(
            f"Unable to find test class {class_name}") from err
    if not issubclass(cls, testbase.TestCase):
        raise TestClassError(
            f"Test class {class_name} not derived from test_testbase.TestCase")
    return cls


def runTests(ctx: context.TestContext,
             tgt: target.TestTarget,
             stop_on_error: bool = False) -> List[testbase.TestResult]:
    """Run all the tests in the TestContext against the TestTarget.

    Args:
        ctx: A context.TestContext object.
        tgt: A target.TestTarget with a connection established.
        stop_on_error: If True, execution stop if there is an error.

    Returns:
        A list of testbase.TestResult objects.

    """
    suites = getTestSuites(ctx, tgt)
    errored_test = len(ctx.tests) - len(suites)
    if errored_test:
        logging.error("Running %d/%d Tests, %d failed to load.",
                      len(suites), len(ctx.tests), errored_test)
    if stop_on_error:
        logging.info("Stopping if a Test fails.")
    results = []
    conn_lock = threading.Lock()
    for i, suite in enumerate(suites):
        thread = TestThread(suite, conn_lock, stop_on_error)
        logging.debug(f"Running Test {i+1} '{suite.test_name}'")
        thread.start()
        thread.join()
        results.append(thread.results)
        logging.info(
            f"Test {i+1} '{suite.test_name}' took "
            f"{thread.results.duration_sec} secs: "
            f"{'' if thread.results.wasSuccessful() else 'NOT '}"
            "PASSED\n")
        if stop_on_error and not thread.results.wasSuccessful():
            break
    return results


def setInitConfigs(ctx: context.TestContext,
                   tgt: target.TestTarget,
                   stop_on_error: bool = False) -> bool:
    """Applies the initial configurations to the target.

    Args:
        ctx: A context.TestContext object.
        tgt: A target.TestTarget with a connection established.
        stop_on_error: If True, stop if there is an error.

    Raises:
        InitConfigError if failed to apply the config. Does not raise if
        stop_on_error.

    """
    for init_config in ctx.init_configs:
        if (not getattr(init_config, "filename", None) or
                not getattr(init_config, "xpath", None)):
            msg = "Initial configuration file and xpath are both needed."
            if stop_on_error:
                raise InitConfigError(msg)
            logging.error(msg)
        try:
            schema.parsePath(init_config.xpath)
            tgt.gNMISetConfigFile(init_config.filename, init_config.xpath)
            logging.info(f"Initial OpenConfig '{init_config.filename}' "
                         f"applied at {init_config.xpath}")
        except (IOError, json.JSONDecodeError, target.BaseError) as err:
            msg = (f"Unable to set configuration '{init_config.filename}' "
                   f"at '{init_config.xpath}': {err}")
            if stop_on_error:
                raise InitConfigError(msg) from err
            logging.error(msg)



# -*- coding: utf-8 -*-
from operator import attrgetter
from pyangbind.lib.yangtypes import RestrictedPrecisionDecimalType
from pyangbind.lib.yangtypes import RestrictedClassType
from pyangbind.lib.yangtypes import TypedListType
from pyangbind.lib.yangtypes import YANGBool
from pyangbind.lib.yangtypes import YANGListType
from pyangbind.lib.yangtypes import YANGDynClass
from pyangbind.lib.yangtypes import ReferenceType
from pyangbind.lib.base import PybindBase
from collections import OrderedDict
from decimal import Decimal
from bitarray import bitarray
import six

# PY3 support of some PY2 keywords (needs improved)
if six.PY3:
  import builtins as __builtin__
  long = int
elif six.PY2:
  import __builtin__

class openconfig_mpls_sr(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-mpls-sr - based on the path /openconfig-mpls-sr. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Configuration for MPLS with segment routing-based LSPs,
including global parameters, and LSP-specific configuration for
both constrained-path and IGP-congruent LSPs
  """
  _pyangbind_elements = {}

  

class openconfig_mpls_ldp(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-mpls-ldp - based on the path /openconfig-mpls-ldp. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Configuration of Label Distribution Protocol global and LSP-
specific parameters for IGP-congruent LSPs.

This model reuses data items defined in the IETF YANG model for
LDP described by draft-ietf-mpls-ldp-yang-04, YANG Data Model for
MPLS LDP, following an alternate structure.

Portions of this code were derived from draft-ietf-mpls-ldp-yang-04.
Please reproduce this note if possible.

IETF code is subject to the following copyright and license:
Copyright (c) IETF Trust and the persons identified as authors of
the code.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, is permitted pursuant to, and subject to the license
terms contained in, the Simplified BSD License set forth in
Section 4.c of the IETF Trust's Legal Provisions Relating
to IETF Documents (http://trustee.ietf.org/license-info).
  """
  _pyangbind_elements = {}

  

class openconfig_mpls_rsvp(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-mpls-rsvp - based on the path /openconfig-mpls-rsvp. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Configuration for RSVP-TE signaling, including global protocol
parameters and LSP-specific configuration for constrained-path
LSPs
  """
  _pyangbind_elements = {}

  

class openconfig_mpls_types(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-mpls-types - based on the path /openconfig-mpls-types. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: General types for MPLS / TE data model
  """
  _pyangbind_elements = {}

  

class openconfig_mpls_te(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-mpls-te - based on the path /openconfig-mpls-te. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Configuration related to constrained-path LSPs and traffic
engineering.  These definitions are not specific to a particular
signaling protocol or mechanism (see related submodules for
signaling protocol-specific configuration).
  """
  _pyangbind_elements = {}

  

class openconfig_mpls_igp(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-mpls-igp - based on the path /openconfig-mpls-igp. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Configuration generic configuration parameters for IGP-congruent
LSPs
  """
  _pyangbind_elements = {}

  

class openconfig_mpls(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-mpls - based on the path /openconfig-mpls. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This module provides data definitions for configuration of
Multiprotocol Label Switching (MPLS) and associated protocols for
signaling and traffic engineering.

RFC 3031: Multiprotocol Label Switching Architecture

The MPLS / TE data model consists of several modules and
submodules as shown below.  The top-level MPLS module describes
the overall framework.  Three types of LSPs are supported:

i) traffic-engineered (or constrained-path)

ii) IGP-congruent (LSPs that follow the IGP path)

iii) static LSPs which are not signaled

The structure of each of these LSP configurations is defined in
corresponding submodules.  Companion modules define the relevant
configuration and operational data specific to key signaling
protocols used in operational practice.


                         +-------+
       +---------------->| MPLS  |<--------------+
       |                 +-------+               |
       |                     ^                   |
       |                     |                   |
  +----+-----+      +--------+-------+     +-----+-----+
  | TE LSPs  |      | IGP-based LSPs |     |static LSPs|
  |          |      |                |     |           |
  +----------+      +----------------+     +-----------+
      ^  ^                    ^  ^
      |  +----------------+   |  +--------+
      |                   |   |           |
      |   +------+      +-+---+-+      +--+--+
      +---+ RSVP |      |SEGMENT|      | LDP |
          +------+      |ROUTING|      +-----+
                        +-------+

  """
  _pyangbind_elements = {}

  

class openconfig_mpls_te(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-mpls-te - based on the path /openconfig-mpls-te. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Configuration related to constrained-path LSPs and traffic
engineering.  These definitions are not specific to a particular
signaling protocol or mechanism (see related submodules for
signaling protocol-specific configuration).
  """
  _pyangbind_elements = {}

  

class openconfig_mpls_igp(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-mpls-igp - based on the path /openconfig-mpls-igp. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Configuration generic configuration parameters for IGP-congruent
LSPs
  """
  _pyangbind_elements = {}

  

class openconfig_mpls_static(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-mpls-static - based on the path /openconfig-mpls-static. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Defines static LSP configuration
  """
  _pyangbind_elements = {}

  

class openconfig_mpls_static(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-mpls-static - based on the path /openconfig-mpls-static. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Defines static LSP configuration
  """
  _pyangbind_elements = {}

  

